function RDF(){
	this._ds = Components.classes["@mozilla.org/rdf/datasource;1?name=in-memory-datasource"]
					.createInstance(Components.interfaces.nsIRDFDataSource);
	this._rdfService = Components.classes['@mozilla.org/rdf/rdf-service;1']
					.getService(Components.interfaces.nsIRDFService);
	this._rdfContainerUtils = Components.classes['@mozilla.org/rdf/container-utils;1']
					.getService(Components.interfaces.nsIRDFContainerUtils);
}


RDF.prototype = {

	get NS_TABTREE() {
		return "http://park2.wakwak.com/~benki/tabtree#";
	},
	
	get RDF_ITEM_ROOT() {
		return "urn:tabtree:item:root";
	},
	
	get RDF_ITEM_SEARCH() {
		return "urn:tabtree:item:search";
	},
	
	get ds(){
		return this._ds;
	},
	
	getResource: function(aResourceID){
		return this._rdfService.GetResource(aResourceID);
	},
	
	getLiteral: function(aValue){
		return this._rdfService.GetLiteral(aValue);
	},
	
	getIntLiteral: function(aValue){
		return this._rdfService.GetIntLiteral(aValue);
	},
	
	getDateLiteral: function(aValue){
		return this._rdfService.GetDateLiteral(aValue);
	},
	
	setLiteralProperty: function(aResourceOrID, aProperty, aValue){
		var source = (typeof(aResourceOrID)=="string") ?
				this.getResource(aResourceOrID) : aResourceOrID;
		var property = this.getResource(aProperty);
		var value = this.getLiteral(aValue);
		
		this.ds.Assert(source, property, value, true);
	},
	
	getLiteralProperty: function(aResourceOrID, aProperty){
		var source = (typeof(aResourceOrID)=="string") ?
						this.getResource(aResourceOrID) : aResourceOrID;
		var target = this.ds.GetTarget(source, this.getResource(aProperty), true);
		if(target) return target.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
		return null;
	},
	
	changeLiteralProperty: function(aResourceOrID, aProperty, aNewValue) {
		try {
			var source = (typeof(aResourceOrID)=="string") ?
					this.getResource(aResourceOrID) : aResourceOrID;
			var property = this.getResource(aProperty);
			var oldValue = this.ds.GetTarget(source, property, true);
			if( !oldValue ) return null;
			//oldValue = oldValue.QueryInterface(Components.interfaces.nsIRDFLiteral).Value;
			var newValue = this.getLiteral(aNewValue);
			
			return this.ds.Change(source, property, oldValue, newValue);
		} catch(e) {
			return this.setLiteralProperty(aResourceOrID, aProperty, aNewValue);
		}
	},
	
	setIntProperty: function(aResourceOrID, aProperty, aValue){
		var source = (typeof(aResourceOrID)=="string") ?
						this.getResource(aResourceOrID) : aResourceOrID;
		var property = this.getResource(aProperty);
		var value = this.getIntLiteral(aValue);

		this.ds.Assert(source, property, value, true);
	},
	
	getIntProperty: function(aResourceOrID, aProperty){
		var source = (typeof(aResourceOrID)=="string") ?
							this.getResource(aResourceOrID) : aResourceOrID;
		var target = this.ds.GetTarget(source, this.getResource(aProperty), true);
		if(target) return target.QueryInterface(Components.interfaces.nsIRDFInt).Value;
		return null;
	},
	
	setDateProperty: function(aResourceOrID, aProperty, aValue){
		var source = (typeof(aResourceOrID)=="string") ?
							this.getResource(aResourceOrID) : aResourceOrID;
		var property = this.getResource(aProperty);
		var value = this.getDateLiteral(aValue);

		this.ds.Assert(source, property, value, true);
	},
	
	getDateProperty: function(aResourceOrID, aProperty){
		var source = (typeof(aResourceOrID)=="string") ?
							this.getResource(aResourceOrID) : aResourceOrID;
		var target = this.ds.GetTarget(source, this.getResource(aProperty), true);
		if(target) return target.QueryInterface(Components.interfaces.nsIRDFDate).Value;
		return null;
	},
	
	appendResource: function(aResourceOrID, aParentContainerOrID){
		var resource = (typeof(aResourceOrID)=="string") ?
							this.getResource(aResourceOrID) : aResourceOrID;
		if(aParentContainerOrID){
			var parentContainer = (typeof(aParentContainerOrID)=="string") ?
				this.getContainer(aParentContainerOrID) : aParentContainerOrID;
			parentContainer.AppendElement(resource);
		}
		return resource;
	},
	
	removeResource: function(aResourceOrID, aParentContainerOrID) {
		var resource = (typeof(aResourceOrID)=="string") ?
							this.getResource(aResourceOrID) : aResourceOrID;
		
		// 親コンテナからリソースを削除
		if( aParentContainerOrID ) {
			var parentContainer = (typeof(aParentContainerOrID)=="string") ?
				this.getContainer(aParentContainerOrID) : aParentContainerOrID;
			parentContainer.RemoveElement(resource, true);
		}
		
		// リソースを削除
		var names = this.ds.ArcLabelsOut(resource);
		var name, value;
		while( names.hasMoreElements() ) {
			name = names.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
			value = this.ds.GetTarget(resource, name, true);
			this.ds.Unassert(resource, name, value);
		}
		
	},
	
	insertResource: function(aResourceOrID, aParentContainerOrID, aIndex) {
		var resource = (typeof(aResourceOrID)=="string") ?
							this.getResource(aResourceOrID) : aResourceOrID;
		var parentContainer = (typeof(aParentContainerOrID)=="string") ?
				this.getContainer(aParentContainerOrID) : aParentContainerOrID;
		parentContainer.InsertElementAt(resource, aIndex, true);
		return resource;
	},
	
	moveResource: function(aOldResourceOrID, aNewResourceOrID) {
		var oldResource = (typeof(aOldResourceOrID)=="string") ?
							this.getResource(aOldResourceOrID) : aOldResourceOrID;
		var newResource = (typeof(aNewResourceOrID)=="string") ?
							this.getResource(aNewResourceOrID) : aNewResourceOrID;
							
		var names = this.ds.ArcLabelsOut(oldResource);
		var name, value;
		while( names.hasMoreElements() ) {
			name = names.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
			value = this.ds.GetTarget(resource, name, true);
			this.ds.Move(oldResource, newResource, name, value);
		}
	},
	
	makeSeqContainer: function(aResourceOrID){
		var containerRes = (typeof(aResourceOrID)=="string") ?
							this.getResource(aResourceOrID) : aResourceOrID;
		return this._rdfContainerUtils.MakeSeq(this.ds, containerRes);
	},
	
	getContainer: function(aResourceOrID){
		try{
			var containerRes = (typeof(aResourceOrID)=="string") ?
							this.getResource(aResourceOrID) : aResourceOrID;
			var rdfContainer = Components.classes["@mozilla.org/rdf/container;1"]
						.createInstance(Components.interfaces.nsIRDFContainer);
			rdfContainer.Init(this.ds, containerRes);
			return rdfContainer;
		}catch(ex){}
		return this.makeSeqContainer(aResourceOrID);
	},
	
	clearContainer: function(aResourceID){
		var container = this.getContainer(aResourceID);
		while(container.GetCount()){
			container.RemoveElementAt(1, true);
		}
	},
	
	search: function(aSearchString){
		this.clearContainer(this.RDF_ITEM_SEARCH);
		var searchContainer = this.getContainer(this.RDF_ITEM_SEARCH);
		
		var list = aSearchString.toUpperCase().replace(/ +/g, " ").replace(/ $/, "").split(" ");
		
		function checkKey(aKey, aList) {
			var length = aList.length;
			var rv = true;
			for (var i=0; i<length; i++) {
				rv = rv && ( ( aKey.toUpperCase().indexOf(aList[i]) > -1 ) ? true : false );
				if (!rv) break;
			}
			return rv;
		}
		
		var resources = this._ds.GetAllResources();
		
		while (resources.hasMoreElements()) {
			var resource = resources.getNext().QueryInterface(Components.interfaces.nsIRDFResource);
			var key = this.getLiteralProperty(resource, this.NS_TABTREE + "title");
			if (key && checkKey(key, list))
				searchContainer.AppendElement(resource);
		}
	},
	
	getXmlSource: function(aNsPrefix, aNsURL){
		var serializer = Components.classes["@mozilla.org/rdf/xml-serializer;1"]
								.createInstance(Components.interfaces.nsIRDFXMLSerializer);
		serializer.init(this.ds);
		
		if(aNsPrefix && aNsURL){
			var atomService = Components.classes["@mozilla.org/atom-service;1"]
									.createInstance(Components.interfaces.nsIAtomService);
			var prefix = atomService.getAtom(aNsPrefix);
			serializer.addNameSpace(prefix, aNsURL);
		}

		var outputStream = {
			data: "",
			write: function(aBuffer, aCount){
				this.data += aBuffer;
				return aCount;
			}
		};

		serializer.QueryInterface(Components.interfaces.nsIRDFXMLSource)
						.Serialize(outputStream);
		return(outputStream.data);
	},
	
	isEmpty: function(aResourceOrID) {
		var resource = (typeof(aResourceOrID)=="string") ?
							this.getResource(aResourceOrID) : aResourceOrID;
		return this._rdfContainerUtils.IsEmpty(this.ds, resource);
	},
	
	isContainer: function(aResourceOrID) {
		var resource = (typeof(aResourceOrID)=="string") ?
							this.getResource(aResourceOrID) : aResourceOrID;
		return this._rdfContainerUtils.IsContainer(this.ds, resource);
	}
	
}
