const Ci = Components.interfaces;
const Cc = Components.classes;
const Cu = Components.utils;

Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/XPCOMUtils.jsm");

let TabTree = {
	
	RDF_ITEM_ROOT:		"urn:tabtree:item:root",
	RDF_ITEM_SEARCH:	"urn:tabtree:item:search",
	NS_TABTREE:			"http://park2.wakwak.com/~benki/tabtree#",
	
	init: function TabTree_init() {
		this.tree = document.getElementById("tabtree-tree");
		if (!this.tree)
			return;
		
		if (!this._tabbrowser)
			return;
			
		let builder = this.tree.builder.QueryInterface(Ci.nsIXULTreeBuilder);
		builder.addObserver(this);
		
		this.initRDF();
		this.initEventListeners();
		
		this.tree.builder.rebuild();
		this.selectCurrentTreeItem();
	},
	
	initRDF: function TabTree_initRDF() {
		this._RDF = new RDF();
		this._RDFRoot = this._RDF.makeSeqContainer(this.RDF_ITEM_ROOT);
		
		let tab = this._tabbrowser.mTabContainer.firstChild;
		while (tab) {
			let opener = tab.getAttribute("opener");
			if (opener) {
				let openerPanel = this.getPanelById(opener);
				if (openerPanel) {
					this.appendResource(this.getPanelById(tab.getAttribute("linkedpanel")), openerPanel);
				} else {
					this.appendResource(this.getPanelById(tab.getAttribute("linkedpanel")), null);
				}
			} else {
				this.appendResource(this.getPanelById(tab.getAttribute("linkedpanel")), null);
			}
			tab = tab.nextSibling;
		}
		this.tree.database.AddDataSource(this._RDF.ds);
		this.tree.setAttribute("ref", this.RDF_ITEM_ROOT);
	},
	
	initEventListeners: function TabTree_initEventListeners() {
		this.tree.addEventListener("click", function(event) { this.onClick(event); }.bind(this), true);
		this.tree.addEventListener("DOMMouseScroll", function(event) { this.onMouseScroll(event); }.bind(this), true);
		
		let tabs = this._topWin.document.getElementById("tabbrowser-tabs");
		this.tabMutationObserver = new MutationObserver(function(mutations) {
			this.onTabMutationObserve(mutations);
		}.bind(this));
		this.tabMutationObserver.observe(tabs, { childList: true, attributes: true });
		
		this._tabbrowser.addEventListener("DOMTitleChanged", this, true);
		this._tabbrowser.addEventListener("click", this, true);
		this._tabbrowser.addEventListener("pageshow", this, true);
		this._tabbrowser.addEventListener("select", this, true);
		this._tabbrowser.addEventListener("DOMLinkAdded", this, false);
		
		let tab = this._tabbrowser.mTabContainer.firstChild;
		while (tab) {
			tab.addEventListener("error", this, true);
			tab.linkedBrowser.webProgress.addProgressListener(this, Ci.nsIWebProgress.NOTIFY_ALL);
			tab = tab.nextSibling;
		}
	},
	
	onTabMutationObserve: function TabTree_onTabMutationObserve(mutations) {
		let self = this;
		mutations.forEach(
			function(mutation) {
				if (mutation.type == "childList") {
					for (let i=0; i<mutation.addedNodes.length; ++i) {
						let tab = mutation.addedNodes[i];
						if (tab.localName != "tab")
							continue;
						//this.log("addtab");
						tab.addEventListener("error", this, true);
						let timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
						timer.initWithCallback({
							notify: function(timer) {
								try {
									tab.linkedBrowser.webProgress.addProgressListener(this, Ci.nsIWebProgress.NOTIFY_ALL);
								} catch(e) {
								}
								if (this._isMiddleClick) {
									tab.setAttribute("opener", this._currentPanel.id);
									this._isMiddleClick = false;
									this._currentPanel = null;
								}
								
								let opener = tab.getAttribute("opener");
								if (opener) {
									let openerPanel = this.getPanelById(opener);
									if (openerPanel)
										this.appendResource(this.getPanelById(tab.getAttribute("linkedpanel")), openerPanel);
									else
										this.appendResource(this.getPanelById(tab.getAttribute("linkedpanel")), null);
								} else {
									this.appendResource(this.getPanelById(tab.getAttribute("linkedpanel")), null);
								}
								this.tree.builder.rebuild();
								this.selectCurrentTreeItem();
							}.bind(this)
						}, 0, Ci.nsITimer.TYPE_ONE_SHOT);
					}
					for (let i=0; i<mutation.removedNodes.length; ++i) {
						let tab = mutation.removedNodes[i];
						if (tab.localName != "tab")
							continue;
						//this.log("removetab");
						tab.removeEventListener("error", this, true);
						//if (tab.linkedBrowser)
						//	tab.linkedBrowser.webProgress.removeProgressListener(this, Ci.nsIWebProgress.NOTIFY_ALL);
						//this.log("mutation");
						let panel = tab.getAttribute("linkedpanel");
						if (this._RDF.isContainer(panel) &&
							!this._RDF.isEmpty(panel)) {
							let container = this._RDF.getContainer(panel);
							let resources = container.GetElements();
							while (resources.hasMoreElements()) {
								let resource = resources.getNext().QueryInterface(Ci.nsIRDFResource);
								this._RDF.appendResource(resource, this._RDFRoot);
							}
							while (container.GetCount()) {
								container.RemoveElementAt(1, true);
							}
						}
						
						let opener = this.getTabById(tab.getAttribute("opener"));
						this._RDF.removeResource(panel, opener ? opener.getAttribute("linkedpanel") : this._RDFRoot);
						
						this.tree.builder.rebuild();
						this.selectCurrentTreeItem();
					}
				}
				/*
				if (mutation.type == "attributes") {
					//this.log(mutation.attributeName);
				}
				*/
			}.bind(this)
		);
	},
	
	finish: function TabTree_finish() {
		if (this.tree) {
			let builder = this.tree.builder.QueryInterface(Ci.nsIXULTreeBuilder);
			builder.removeObserver(this);
		}
				
		this.clearEventListeners();
		
		this._tabbrowser = null;
		this._contextTab = null;
		this._currentPanel = null;
		this._RDF = null;
		this._RDFRoot = null;
	},
	
	clearEventListeners: function TabTree_clearEventListeners() {
		this.tree.removeEventListener("click", function(event) { this.onClick(event); }.bind(this), true);
		this.tree.removeEventListener("DOMMouseScroll", function(event) { this.onMouseScroll(event); }.bind(this), true);
		
		//let tabs = this._topWin.document.getElementById("tabbrowser-tabs");
		this._tabbrowser.removeEventListener("DOMTitleChanged", this, true);
		this._tabbrowser.removeEventListener("click", this, true);
		this._tabbrowser.removeEventListener("pageshow", this, true);
		this._tabbrowser.removeEventListener("select", this, true);
		this._tabbrowser.removeEventListener("DOMLinkAdded", this, false);
		
		this.tabMutationObserver.disconnect();
		
		let tab = this._tabbrowser.mTabContainer.firstChild;
		while (tab) {
			tab.removeEventListener("error", this, true);
			tab.linkedBrowser.webProgress.removeProgressListener(this, Ci.nsIWebProgress.NOTIFY_ALL);
			tab = tab.nextSibling;
		}
	},
	
	log: function TabTree_log(aMessage) {
		Services.console.logStringMessage(aMessage);
	},
	
	appendResource: function TabTree_appendResource(aPanel, aOpenerPanel) {
		let container = aOpenerPanel ?
							this._RDF.getContainer(aOpenerPanel.id) :
							this._RDFRoot;
		let resource = this._RDF.appendResource(aPanel.id, container);
		this._RDF.getContainer(aPanel.id)
		let title = this.getBrowserTitle(aPanel);
		this._RDF.setLiteralProperty(resource, this.NS_TABTREE+"title", title);
		this._RDF.setLiteralProperty(resource, this.NS_TABTREE+"id", aPanel.id);
		
		let browser = this.getContentBrowser(aPanel);
		if (browser.currentURI.scheme == "bbs2ch")
			this._RDF.setLiteralProperty(resource, this.NS_TABTREE+"type", "bbs2ch");
		else if (browser.contentDocument instanceof ImageDocument)
			this._RDF.setLiteralProperty(resource, this.NS_TABTREE+"type", "image");
		else
			this._RDF.setLiteralProperty(resource, this.NS_TABTREE+"type", "normal");
		
		let tab = this.getTabById(this._tabbrowser.getNotificationBox(browser).id);
		//dump(browser.mIconURL+"\n");
		if (tab.hasAttribute("busy")) {
			this._RDF.setLiteralProperty(resource, this.NS_TABTREE+"icon", "chrome://tabtree/skin/connecting.png");
		} else if (!tab.hasAttribute("image")) {
			this._RDF.setLiteralProperty(resource, this.NS_TABTREE+"icon", "chrome://tabtree/skin/defaultFavicon.png");
		} else if (browser.mIconURL) {
			let fs = Cc["@mozilla.org/browser/favicon-service;1"].getService(Ci.nsIFaviconService);
			if (fs) {
				let uri = Services.io.newURI(browser.mIconURL, null, null);
				//dump(fs.isFailedFavicon(uri)+"\n");
				if (!fs.isFailedFavicon(uri))
					this._RDF.setLiteralProperty(resource, this.NS_TABTREE+"icon", browser.mIconURL);
			}
			
		} else {
			this._RDF.setLiteralProperty(resource, this.NS_TABTREE+"icon", "chrome://tabtree/skin/defaultFavicon.png");
		}
		
		//コンテナが閉じてたら開く
		if (aOpenerPanel) {
			let index = this.tree.builder.getIndexOfResource(this._RDF.getResource(aOpenerPanel.id));
			if (index != -1) {
				if (!this.tree.view.isContainerOpen(index)) {
					this.tree.view.toggleOpenState(index);
				}
			}
		}
	},
	
	checkTabs: function TabTree_checkTabs() {
		let ids = [];
		this.getAllChildren(this._RDFRoot, ids);
		if (ids.length != this._tabbrowser.mTabContainer.childNodes.length) {
			let tab;
			for (let i=0; i<ids.length; i++) {
				tab = this.getTabById(ids[i]);
				if (!tab) {
					let resource = this._RDF.getResource(ids[i]);
					let index = this.tree.view.getIndexOfResource(resource);
					index = this.tree.view.getParentIndex(index);
					let container = ( index >= 0 ) ? this._RDF.getContainer(this.tree.view.getResourceAtIndex(index)): this._RDFRoot;
					this._RDF.removeResource(resource, container);
				}
			}
		}
	},
	
	selectCurrentTreeItem: function TabTree_selectCurrentTreeItem(aPanel) {
		let browser;
		if (aPanel)
			browser = this.getContentBrowser(aPanel._selectedPanel);
		else
			browser = this._tabbrowser.mCurrentBrowser;
		
		if (!browser) return;
		
		let index = this.tree.builder.getIndexOfResource(this._RDF.getResource(browser.parentNode.parentNode.parentNode.parentNode.id));
		this.tree.view.selection.select(index);
		this.tree.treeBoxObject.ensureRowIsVisible(index);
	},
	
	updatePopupMenu: function TabTree_updatePopupMenu(aPopupMenu, aEvent) {
		let row = new Object();
		let obj = new Object();
		this.tree.treeBoxObject.getCellAt(aEvent.clientX, aEvent.clientY, row, {}, obj);
		
		if (row.value == -1) {
			this._contextTab = aPopupMenu; //とりあえず
		} else {
			let resource = this.tree.view.getResourceAtIndex(row.value);
			let id = this._RDF.getLiteralProperty(resource, this.NS_TABTREE+"id");
			
			let tab = this.getTabById(id);
			if (!tab)
				this._contextTab = aPopupMenu;
			else
				this._contextTab = tab;
		}
		
		let disabled = this._tabbrowser.mPanelContainer.childNodes.length == 1;
		let menuItems = aPopupMenu.getElementsByAttribute("tabattr", "tabbrowser-multiple");
		for(let i=0; i<menuItems.length; i++) {
			if (disabled)
				menuItems[i].setAttribute("disabled", disabled);
			else
				if (menuItems[i].hasAttribute("disabled"))
					menuItems[i].removeAttribute("disabled");
		}
	},
	
	closeTree: function TabTree_closeTree(aTab, aCloseParent) {
		if (aTab.localName != "tab")
			aTab = this._tabbrowser.mCurrentTab;
		else
			this._tabbrowser.mTabContainer.selectedItem = aTab;
		
		if (this._RDF.isContainer(this._RDF.getResource(aTab.linkedPanel))) {
			let container = this._RDF.getContainer(aTab.linkedPanel);
			let ids = [];
			this.getAllChildren(container, ids);
			//for (let i=ids.length-1; i>=0; i--) {
			for (let i=0; i<ids.length; ++i) {
				let tab = this.getTabById(ids[i]);
				if (tab && ( tab != aTab )) {
					this.log("removetab");
					this._tabbrowser.removeTab(tab);
				}
			}
		}
		
		// selectedItem を removeTab するとダメっぽい。
		// updateCurrentBrowser() で setTimeout してるから、
		// ちょっと待ってから。
		if (aCloseParent) {
			let timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
			timer.initWithCallback({
				notify: function(timer) {
					let length = this._tabbrowser.mTabContainer.childNodes.length;
					if (length == 1) {
						this._tabbrowser.removeCurrentTab();
						return;
					}
					
					for(let i=0; i<length; i++) {
						if (aTab != this._tabbrowser.mTabContainer.childNodes[i]) {
							this._tabbrowser.mTabContainer.selectedItem = this._tabbrowser.mTabContainer.childNodes[i];
							let timer2 = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
							timer2.initWithCallback({
								notify: function(timer) {
									this._tabbrowser.removeTab(aTab);
								}.bind(this)
							}, 0,  Ci.nsITimer.TYPE_ONE_SHOT);
							break;
						}
					}
					
				}.bind(this)
			}, 0, Ci.nsITimer.TYPE_ONE_SHOT);
			//setTimeout(removeATab, 0);
		}
	},
	
	searchTabs: function TabTree_searchTabs(aSearchString) {
		aSearchString = aSearchString.replace(/^ +/, "");
		if (aSearchString) {
			this._RDF.search(aSearchString);
			this.tree.ref = this.RDF_ITEM_SEARCH;
		} else {
			this.tree.ref = this.RDF_ITEM_ROOT;
		}
		this.selectCurrentTreeItem();
	},
	
	onClick: function TabTree_onClick(event) {
		let row = new Object();
		let col = new Object();
		let obj = new Object();
		this.tree.treeBoxObject.getCellAt(event.clientX, event.clientY, row, col, obj);
		
		if (obj.value == "twisty") return; //twistyクリックのときは何もしない
		
		let eButton = event.button;
		let eDetail = event.detail;
		if (row.value == -1 && eDetail == 2 && eButton == 0) { //アイテム以外をダブルクリックのとき
			let newTab = this._tabbrowser.addTab("about:blank");
			let background = Services.prefs.getBoolPref("browser.tabs.loadInBackground");
			if (!background)
				this._tabbrowser.selectedTab = newTab;
			event.preventDefault();
			event.stopPropagation();
			return;
		} else if (row.value == -1) { //アイテム以外をクリックのとき
			return;
		}
		
		let resource = this.tree.view.getResourceAtIndex(row.value);
		let id = this._RDF.getLiteralProperty(resource, this.NS_TABTREE+"id");
		
		let tab = this.getTabById(id);
		if (!tab) return;
		
		if (eButton == 0) //左クリック
			this._tabbrowser.selectedTab = tab;
		else if ( eButton == 1 ) //中クリック
			this._tabbrowser.removeTab(tab);
	},
	
	onMouseScroll: function TabTree_onMouseScroll(event) {
		let index = this.tree.currentIndex;
		let count = this.tree.view.rowCount;
		
		(event.detail > 0) ? index++ : index-- ;
		if (index < 0)
			index = count-1;
		else if (index > count-1)
			index = 0;
		
		let resource = this.tree.view.getResourceAtIndex(index);
		let id = this._RDF.getLiteralProperty(resource, this.NS_TABTREE+"id");
		
		let tab = this.getTabById(id);
		if (tab) {
			this._tabbrowser.selectedTab = tab;
		}
		
		this.tree.view.selection.select(index);
		this.tree.treeBoxObject.ensureRowIsVisible(index);
	},
	
	getContentBrowser: function TabTree_getContentBrowser(aPanel) {
		let tab = this.getTabById(aPanel.id);
		return tab.linkedBrowser;
	},
	
	getPanelById: function TabTree_getPanelById(aID) {
		return this._topWin.document.getAnonymousElementByAttribute(this._tabbrowser, "id", aID);
	},
	
	getTabById: function TabTree_getTabById(aID) {
		return this._topWin.document.getElementsByAttribute("linkedpanel", aID).item(0);
	},
	
	getBrowserTitle: function TabTree_getBrowserTitle(aPanel) {
		let browser = this.getContentBrowser(aPanel);
		let title = browser.contentTitle;
		if (title) return title;
		
		let uri = browser.currentURI;
		if (uri && (uri.spec != "about:blank")) return uri.spec;
		
		return this._tabbrowser.mStringBundle.getString("tabs.emptyTabTitle");
	},
	
	getAllChildren: function TabTree_getAllChildren(aContainer, aArray) {
		let children = aContainer.GetElements();
		while (children.hasMoreElements()) {
			let child = children.getNext();
			if (child instanceof Ci.nsIRDFResource) {
				let resource = child.QueryInterface(Ci.nsIRDFResource);
				let id = this._RDF.getLiteralProperty(resource, this.NS_TABTREE+"id");
				aArray.push(id);
				if (this._RDF.isContainer(resource))
					this.getAllChildren(this._RDF.getContainer(resource), aArray);
			}
		}
	},
	
	get _tabbrowser() {
		delete this._tabbrowser;
		return this._tabbrowser = this._topWin.document.getElementById("content");
	},
	
	get _topWin() {
		delete this._topWin;
		return this._topWin = Services.wm.getMostRecentWindow("navigator:browser");
	},
	
	/***** ***** nsIXULTreeBuilderObserver ***** *****/
	
	canDrop: function TabTree_canDrop(aIndex, aOrientation, aDataTransfer) {
		let dragSession = this._dragService.getCurrentSession();
		if (!dragSession)
			return false;
		
		return true;
	},
	
	onDrop: function TabTree_onDrop(aRow, aOrientation, aDataTransfer) {
		let dragSession = this._dragService.getCurrentSession();
		if (!dragSession)
			return;
		
		let trans = Cc["@mozilla.org/widget/transferable;1"].createInstance(Ci.nsITransferable);
		trans.addDataFlavor("tabtree/item");
		trans.addDataFlavor("text/x-moz-url");
		trans.addDataFlavor("text/unicode");
		
		let url, extra, item, parent;
		let bestFlavour = {}, dataObj = {}, len = {};
				
		dragSession.getData(trans, 0);
		trans.getAnyTransferData(bestFlavour, dataObj, len);
		dataObj = dataObj.value.QueryInterface(Ci.nsISupportsString);
		dataObj = dataObj.data.substring(0, len.value).split("\n");
		url = dataObj[0];
		if (dataObj.length > 1 && dataObj[1] != "")
			extra = dataObj[1];
		else
			extra = null;
				
		let target = {};
		if (aRow >= 0) {
			target.item = this.tree.view.getResourceAtIndex(aRow);
			let index = this.tree.view.getParentIndex(aRow);
			target.parent = (index >= 0) ? this.tree.view.getResourceAtIndex(index): this._RDF.getResource(this.RDF_ITEM_ROOT);
		} else {
			target.item = null;
			target.parent = this._RDF.getResource(this.RDF_ITEM_ROOT);
		}
		
		switch (bestFlavour.value) {
			case "tabtree/item":
				item = this._RDF.getResource(url);
				parent = extra ? this._RDF.getResource(extra) : this._RDF.getResource(this.RDF_ITEM_ROOT);
				break;
			case "text/x-moz-url":
			case "text/unicode":
				break;
		}
		
		this.tree.view.selection.selectEventsSuppressed = true;
		
		let self = this;
		function rebuild() {
			self.tree.builder.rebuild();
			self.selectCurrentTreeItem();
			let timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
			timer.initWithCallback({
				notify: function(timer) {
					self.tree.view.selection.selectEventsSuppressed = false;
				}
			}, 100, Ci.nsITimer.TYPE_ONE_SHOT);
		}
		
		if (dragSession.isDataFlavorSupported("tabtree/item")) {
			if (!target.item) {
				this._RDF.getContainer(parent).RemoveElement(item, true);
				this._RDF.appendResource(item, this._RDFRoot);
				this.getTabById(item.Value).removeAttribute("opener");
				rebuild();
				return;
			}
			
			if ((item.Value == target.parent.Value) ||
				(item.Value == target.item.Value)) {
				rebuild();
				return;
			}
			
			if (aOrientation == 0) {
				let targetContainer = this._RDF.getContainer(target.item);
				this._RDF.appendResource(item, targetContainer);
				this._RDF.getContainer(parent).RemoveElement(item, true);
				
				this.getTabById(item.Value).setAttribute("opener", target.item.Value);
			} else {
				let targetContainer = this._RDF.getContainer(target.parent);
				let targetIndex = targetContainer.IndexOf(target.item);
				let container = this._RDF.getContainer(parent);
				let index = container.IndexOf(item);
				container.RemoveElementAt(index, true);
				targetContainer.InsertElementAt(item, targetIndex, true);
				
				if (target.parent.Value == this.RDF_ITEM_ROOT)
					this.getTabById(item.Value).removeAttribute("opener");
				else
					this.getTabById(item.Value).setAttribute("opener", target.parent.Value);
			}
		} else if (dragSession.isDataFlavorSupported("text/x-moz-url")) {
			if (!target.item) {
				this._tabbrowser.selectedTab = this._tabbrowser.addTab(url);
				rebuild();
				return;
			}
			
			if (aOrientation == 0) {
				this._isMiddleClick = true;
				this._currentPanel = this.getPanelById(target.item.Value);
				this._tabbrowser.selectedTab = this._tabbrowser.addTab(url);
			} else {
				if (target.parent.Value != this.RDF_ITEM_ROOT) {
					this._isMiddleClick = true;
					this._currentPanel = this.getPanelById(target.parent.Value);
				}
				this._tabbrowser.selectedTab = this._tabbrowser.addTab(url);
			}
		}
		rebuild();
	},
	
	onToggleOpenState: function TabTree_onToggleOpenState(aRow) {
		//dump("onToggleOpenState\n");
		//rebuild();
		let timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
		timer.initWithCallback({
			notify: function(timer) {
				this.selectCurrentTreeItem();
			}.bind(this)
		}, 0, Ci.nsITimer.TYPE_ONE_SHOT);
	},
	
	onCycleHeader: function(aColumnID, aHeaderElement) {},
	onSelectionChanged: function() {},
	onCycleCell: function(aItemIndex, aColumnID) {},
	onPerformAction: function(aAction) {},
	onPerformActionOnRow: function(aAction, aItemIndex) {},
	onPerformActionOnCell: function(aAction, aItemIndex, aColumnID) {},
	
	get _dragService() {
		delete this._dragService;
		return this._dragService = Cc["@mozilla.org/widget/dragservice;1"].getService(Ci.nsIDragService);
		//return Cc["@mozilla.org/widget/dragservice;1"].getService(Ci.nsIDragService);
	},
	
	/***** ***** nsIWebProgressListener ***** *****/
	
	onProgressChange: function(aWebProgress, aRequest,
								aCurSelfProgress, aMaxSelfProgress,
								aCurTotalProgress, aMaxTotalProgress) {},
	
	onStateChange: function TabTree_onStateChange(aWebProgress, aRequest, aStateFlags, aStatus) {
		//dump("onStateChange\n");
		const nsIWebProgressListener = Ci.nsIWebProgressListener;
		
		if (aStateFlags & nsIWebProgressListener.STATE_STOP) {
			const NS_ERROR_UNKNOWN_HOST = 2152398878;
			if (aStatus == NS_ERROR_UNKNOWN_HOST) {
				//dump("NS_ERROR_UNKNOWN_HOST\n");
				return;
			}
		}
		let index = this._tabbrowser.getBrowserIndexForDocument(aWebProgress.DOMWindow.document.defaultView.top.document);
		if (index < 0) return;
		let browser = this._tabbrowser.getBrowserAtIndex(index);
		let resource = this._RDF.getResource(this._tabbrowser.getNotificationBox(browser).id);
		
		if (aStateFlags & nsIWebProgressListener.STATE_START &&
			aStateFlags & nsIWebProgressListener.STATE_IS_NETWORK) {
			
			this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"title", "Loading...");
			this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"icon", "chrome://tabtree/skin/connecting.png");
			//this.tree.builder.rebuild();
		} else if (aStateFlags & nsIWebProgressListener.STATE_STOP &&
				   aStateFlags & nsIWebProgressListener.STATE_IS_NETWORK) {
			//dump("onStateStop\n");
			//dump(browser.contentDocument.title+"\n");
			if (browser.contentDocument instanceof ImageDocument) {
				this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"icon", "chrome://tabtree/skin/image.png");
			} else if (this._tabbrowser.shouldLoadFavIcon(browser.currentURI)) {
				if (!browser.mIconURL) {
					let url = browser.currentURI.prePath + "/favicon.ico";
					if (this._tabbrowser.isFailedIcon) {
						if (!this._tabbrowser.isFailedIcon(url))
							this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"icon", url);
						else
							this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"icon", "chrome://tabtree/skin/defaultFavicon.png");
					} else {
						if (!this._tabbrowser.isIconKnownMissing(url))
							this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"icon", url);
						else
							this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"icon", "chrome://tabtree/skin/defaultFavicon.png");
					}
				} else {
					this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"icon", browser.mIconURL);
				}
			} else {
				this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"icon", "chrome://tabtree/skin/defaultFavicon.png");
			}
			this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"title", this.getBrowserTitle(this._tabbrowser.getNotificationBox(browser)));
			//this.tree.builder.rebuild();
			//this.tree.view.selection.select(this.tree.builder.getIndexOfResource(this._RDF.getResource(browser.parentNode.id)));
			this.selectCurrentTreeItem();
		}
	},
	
	onLocationChange: function(aWebProgress, aRequest, aLocation, aFlags) {},
	onStatusChange: function(aWebProgress, aRequest, aStatus, aMessage) {},
	onSecurityChange: function(aWebProgress, aRequest, aState) {},
	
	/***** ***** nsIEventListener ***** *****/
	
	handleEvent: function TabTree_handleEvent(aEvent) {
		switch (aEvent.type) {
			case "DOMTitleChanged": {
				let tab = this.getTabById(this._tabbrowser.getNotificationBox(this._tabbrowser.getBrowserForDocument(aEvent.originalTarget)).id);
				if (tab) {
					let resource = this._RDF.getResource(tab.getAttribute("linkedpanel"));
					let title = this.getBrowserTitle(this.getPanelById(tab.getAttribute("linkedpanel")));
					this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"title", title);
					
					/*
					let browser = tab.linkedBrowser;
					if (browser.currentURI.scheme == "bbs2ch")
						this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"type", "bbs2ch");
					else if ( browser.contentDocument instanceof ImageDocument )
						this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"type", "image");
					else
						this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"type", "normal");
					*/
				}
				this.tree.builder.rebuild();
				this.selectCurrentTreeItem();
				break;
			}
			case "click": {
				if (aEvent.button == 2) return;
				
				let event = new XPCNativeWrapper(aEvent);
				let target = event.originalTarget;
				while (target) {
					if ( target instanceof HTMLAnchorElement ) break;
					target = target.parentNode;
				}
				
				if (target) {
					if ( (event.button == 1) || (target.getAttribute("target") == "_blank") ) {
						this._isMiddleClick = true;
						this._currentPanel = this._tabbrowser.getNotificationBox(this._tabbrowser.mCurrentBrowser);
					}
				}
				break;
			}
			case "pageshow": {
				//this.tree.builder.rebuild();
				this.selectCurrentTreeItem();
				break;
			}
			case "select": {
				//this.tree.builder.rebuild();
				this.selectCurrentTreeItem(aEvent.originalTarget);
				break;
			}
			case "DOMLinkAdded": {
				//dump("DOMLinkAdded\n");
				//this.log("DOMLinkAdded");
				if (!this._tabbrowser.mPrefs.getBoolPref("browser.chrome.site_icons"))
					return;
				
				if (!aEvent.originalTarget.rel.match((/(?:^|\s)icon(?:\s|$)/i)))
					return;
				
				let href = aEvent.originalTarget.href;
				if (!href)
					return;
					
				const nsIContentPolicy = Ci.nsIContentPolicy;
				try {
							let contentPolicy = Cc['@mozilla.org/layout/content-policy;1'].getService(nsIContentPolicy);
				} catch(e) {
					return;
				}
				
				const secMan = Cc["@mozilla.org/scriptsecuritymanager;1"].getService(Ci.nsIScriptSecurityManager);
				const ioService = Cc["@mozilla.org/network/io-service;1"].getService(Ci.nsIIOService);
				
				const targetDoc = aEvent.target.ownerDocument;
				let uri = ioService.newURI(href, targetDoc.characterSet, null);
				let origURI = ioService.newURI(targetDoc.documentURI, targetDoc.characterSet, null);
				
				const nsIScriptSecMan = Ci.nsIScriptSecurityManager;
				
				const aboutNeterr = "about:neterror?";
				if (origURI.spec.substr(0, aboutNeterr.length) != aboutNeterr)
					return;
				
				try {
					if (!uri.schemeIs("chrome"))
						secMan.checkLoadURI(origURI, uri, nsIScriptSecMan.DISALLOW_SCRIPT);
				} catch(e) {
					return;
				}
				
				if (contentPolicy.shouldLoad(nsIContentPolicy.TYPE_IMAGE,
					uri, origURI, aEvent.target,
					aEvent.target.type,
					null) != nsIContentPolicy.ACCEPT)
					return;
				
				let index = this._tabbrowser.getBrowserIndexForDocument(targetDoc);
				if (index == -1)
					return;
				
				let browser = this._tabbrowser.getBrowserAtIndex(index);
				let resource = this._RDF.getResource(this._tabbrowser.getNotificationBox(browser).id);
				this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"icon", href);
				
				//this.tree.builder.rebuild();
				this.selectCurrentTreeItem();
				
				break;
			}
			case "error": {
				//dump("onerror\n");
				let resource = this._RDF.getResource(aEvent.target.getAttribute("linkedpanel"));
				this._RDF.changeLiteralProperty(resource, this.NS_TABTREE+"icon", "chrome://tabtree/skin/defaultFavicon.png");
				//this.tree.builder.rebuild();
				this.selectCurrentTreeItem();
				break;
			}
		}
	},
	
	/***** ***** DNDObserver ***** *****/
	
	DNDObserver: {
		
		onDragStart: function(aEvent, aXferData, aDragAction) {
			let index = TabTree.tree.view.selection.currentIndex;
			let item = TabTree.tree.view.getResourceAtIndex(index);
			index = TabTree.tree.view.getParentIndex(index);
			let parent = (index >= 0) ? TabTree.tree.view.getResourceAtIndex(index): null;
			
			let url = TabTree.getContentBrowser(TabTree.getPanelById(item.Value)).currentURI.spec;
			let name = TabTree._RDF.getLiteralProperty(item, TabTree.NS_TABTREE+"title");
			
			//this.log(item.Value);
			aXferData.data = new TransferData();
			aXferData.data.addDataForFlavour("tabtree/item", item.Value+"\n"+(parent?parent.Value:""));
			aXferData.data.addDataForFlavour("text/x-moz-url", url+"\n"+name);
			aXferData.data.addDataForFlavour("text/html", "<A HREF='"+url+"'>"+name+"</A>");
			aXferData.data.addDataForFlavour("text/unicode", url);
		},
		
		onDragOver: function(aEvent, aFlavour, aDragSession) {
			if (aDragSession.sourceNode == TabTree.tree)
				aDragSession.canDrop = true;
		},
		
		onDrop: function(aEvent, aXferData, aDragSession) {
		},
		
		getSupportedFlavours: function() {
			let flavourSet = new FlavourSet();
			flavourSet.appendFlavour("tabtree/item");
			flavourSet.appendFlavour("text/x-moz-url");
			flavourSet.appendFlavour("text/unicode");
			return flavourSet;
		}
	},
	
	QueryInterface: XPCOMUtils.generateQI([
						Ci.nsIWebProgressListener,
						Ci.nsIXULTreeBuilderObserver,
						Ci.nsIEventListener,
						Ci.nsISupportsWeakReference,
						Ci.nsISupports
					])
};

//window.addEventListener("load", TabTree.init, false);
//window.addEventListener("unload", TabTree.finish, false);